#define _XOPEN_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <mpi.h>

typedef struct
{
	size_t rows;  // le nombre de lignes
	size_t cols;  // le nombre de colonnes
	double *data; // les valeurs
} matrix_t;

/*****************************************************************************
 *                          Initialize empty matrix                          *
 *****************************************************************************/
void matrix_empty(matrix_t *mat, size_t height, size_t width)
{
	mat->rows = height;
	mat->cols = width;
	mat->data = (double*) calloc(width*height, sizeof(double));
}

/*****************************************************************************
 *                         Initialize random matrix                          *
 *****************************************************************************/
void matrix_random(matrix_t *mat, size_t height, size_t width)
{
	matrix_empty(mat, height, width);
	for (size_t i=0; i<width*height; ++i) mat->data[i] = drand48();
}

/*****************************************************************************
 *                           Delete matrix content                           *
 *****************************************************************************/
void matrix_delete(matrix_t *mat)
{
	mat->cols = 0;
	mat->rows = 0;
	free(mat->data);
}

/*****************************************************************************
 *                               Dispay matrix                               *
 *****************************************************************************/
void matrix_display(matrix_t A)
{
	printf("      ");
	for(size_t j=0; j<A.cols; ++j) printf("   %.3lu   ",j);
	printf("\n    __");
	for(size_t j=0; j<A.cols; ++j) printf("_________");
	printf("_\n");
	for(size_t i=0; i<A.rows; ++i)
	{
		printf("%.3lu | ", i);
		for(size_t j=0; j<A.cols; ++j) printf("%.6f ",A.data[i*A.cols+j]);
		printf("|\n");
	}
	printf("    --");
	for(size_t j=0; j<A.cols; ++j) printf("---------");
	printf("-\n");
}

/*****************************************************************************
 *                              Matrix mutliply                              *
 *****************************************************************************/
void matrix_multiply(matrix_t A, matrix_t B, matrix_t C)
{
	if ((A.rows != C.rows) || (A.cols != B.rows) || (B.cols != C.cols))
	{
		fprintf(stderr, "[ERROR] Incompatible matrix dimensions\n");
		exit(1);
	}
	size_t M = A.rows;
	size_t K = A.cols;
	size_t N = B.cols;

	// Can be used to reset C (not to do for iteratives accumulations)
	// memset(C.data, 0, C.cols * C.rows * sizeof(double));

	for (size_t i=0; i<M; ++i)
	for (size_t j=0; j<N; ++j)
	for (size_t k=0; k<K; ++k)
		C.data[i*N+j] += A.data[i*K+k] * B.data[k*N+j];
}





int main()
{
	srand48(time(NULL));

	matrix_t A;
	matrix_t B;
	matrix_t C;

	matrix_random(&A, 2, 1);
	matrix_random(&B, 1, 3);
	matrix_empty(&C, 2, 3);

	matrix_multiply(A,B,C);

	matrix_display(A);
	matrix_display(B);
	matrix_display(C);

	matrix_delete(&A);
	matrix_delete(&B);
	matrix_delete(&C);

	return 0;
}